#include "Engine.h"
#include "Entitys/Object.h"
#include "Renderer.h"
#include "components/SpriteComponent.h"
#include "utils/FileDialog.h"

#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include <imgui.h>
#include <imgui_impl_glfw.h>
#include <imgui_impl_opengl3.h>


#include <functional>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <vector>
#include <memory>
#include <filesystem>
#include <iostream>
#define WIN32_LEAN_AND_MEAN
#define NOMINMAX
#include <windows.h>
#include <commdlg.h>
#include <openssl/sha.h>

static std::vector<std::shared_ptr<Object>> objects;
static std::shared_ptr<Object> selected = nullptr;
static bool playing = false;
GLFWwindow *window = nullptr;

Engine::Engine()
{
    Init();
}

Engine::~Engine()
{
    Shutdown();
}

void Engine::Init()
{
    glfwInit();
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

    window = glfwCreateWindow(1280, 720, "2D Editor", nullptr, nullptr);
    glfwMakeContextCurrent(window);
    glfwSwapInterval(1);
    glewInit();

    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO &io = ImGui::GetIO();
    io.ConfigFlags |= ImGuiConfigFlags_DockingEnable;
    ImGui::StyleColorsDark();
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init("#version 330");

    Renderer::Init();

    auto obj = std::make_shared<Object>("Sprite Object");
    objects.push_back(obj);
    selected = obj;
}

void Engine::Run()
{
    while (!glfwWindowShouldClose(window))
    {
        glfwPollEvents();

        ImGui_ImplOpenGL3_NewFrame();
        ImGui_ImplGlfw_NewFrame();
        ImGui::NewFrame();
        ImGui::DockSpaceOverViewport(ImGui::GetMainViewport()->ID);

        if (ImGui::BeginMainMenuBar())
        {
            if (ImGui::Button(playing ? "Stop" : "Play"))
            {
                playing = !playing;
            }
            if (ImGui::BeginMenu("File"))
            {
                if (ImGui::MenuItem("Save Scene"))
                {
                    std::string file = OpenFileDialog(FileDialogType::Scenes); // reuse your dialog
                    if (!file.empty())
                        SaveScene(file);
                }
                if (ImGui::MenuItem("Load Scene"))
                {
                    std::string file = OpenFileDialog(FileDialogType::Scenes);
                    if (!file.empty())
                        LoadScene(file);
                }
                ImGui::EndMenu();
            }

            ImGui::EndMainMenuBar();
        }

        ImGui::Begin("Scene Tree");

        

        if (ImGui::BeginPopupContextWindow("SceneTreeContext", ImGuiPopupFlags_MouseButtonRight))
        {
            if (ImGui::MenuItem("Create New Object"))
            {
                auto obj = std::make_shared<Object>("NewObject");
                obj->AddComponent<SpriteComponent>();
                objects.push_back(obj);
                selected = obj;
                ImGui::OpenPopup("RenameObject");
            }
            ImGui::Separator();
            if (ImGui::MenuItem("Create New Sprite Object"))
            {
                auto obj = std::make_shared<Object>("NewSprite");
                obj->AddComponent<SpriteComponent>();
                objects.push_back(obj);
                selected = obj;
                ImGui::OpenPopup("RenameObject");
            }
            ImGui::EndPopup();
        }
        
        


        for (auto &obj : objects)
            if (!obj->GetParent()) // Only draw root nodes
                DrawObjectNode(obj);

        

        ImGui::End();

        ImGui::Begin("Inspector");
        if (selected)
        {
            char buffer[128];
            strcpy(buffer, selected->GetName().c_str());
            if (ImGui::InputText("Name", buffer, sizeof(buffer)))
                selected->SetName(buffer);

            glm::vec2 pos = selected->GetLocalPosition();
            if (ImGui::DragFloat2("Position", &pos.x))
                selected->SetLocalPosition(pos);

            ImGui::InputInt("Layer", &selected->layer);

            // Add component
            if (ImGui::Button("Add SpriteComponent"))
            {
                if (!selected->GetComponent<SpriteComponent>())
                    selected->AddComponent<SpriteComponent>();
            }

            // Show SpriteComponent UI
            if (auto sprite = selected->GetComponent<SpriteComponent>())
            {
                ImGui::Separator();
                ImGui::Text("Sprite Component");

                std::string tex = sprite->GetTexturePath();
                std::string norm = sprite->GetNormalMapPath();

                ImGui::Text("Texture: %s", tex.c_str());
                if (ImGui::Button("Load Texture"))
                {
                    auto path = OpenFileDialog(FileDialogType::Images);
                    if (!path.empty())
                        sprite->SetTexture(path);
                }

                ImGui::Text("Normal Map: %s", norm.c_str());
                if (ImGui::Button("Load Normal Map"))
                {
                    auto path = OpenFileDialog(FileDialogType::Images);
                    if (!path.empty())
                        sprite->SetNormalMap(path);
                }

                if (ImGui::Button("Remove SpriteComponent"))
                {
                    selected->RemoveComponent<SpriteComponent>();
                }
            }
        }
        else
        {
            ImGui::Text("No object selected.");
        }
        ImGui::End();


        
        

        // Viewport
        ImGui::Begin("Viewport", nullptr, ImGuiWindowFlags_NoScrollbar | ImGuiWindowFlags_NoScrollWithMouse);
        ImVec2 size = ImGui::GetContentRegionAvail();
        Renderer::Resize((int)size.x, (int)size.y);
        Renderer::Begin();

        std::vector<std::shared_ptr<Object>> toDraw;

        std::function<void(const std::shared_ptr<Object>&)> collect = [&](const std::shared_ptr<Object>& obj) {
            toDraw.push_back(obj);
            for (const auto& child : obj->GetChildren())
                collect(child);
        };
        
        for (const auto& obj : objects)
            if (!obj->GetParent())
                collect(obj);
        
        std::sort(toDraw.begin(), toDraw.end(), [](const auto& a, const auto& b) {
            return a->layer < b->layer;
        });
        
        for (const auto& obj : toDraw) {
            if (auto sprite = obj->GetComponent<SpriteComponent>()) {
                Renderer::DrawSprite(sprite.get(), obj->GetWorldPosition());
            }
        }
        

        Renderer::End();
        GLuint texID = Renderer::GetRenderTexture();
        ImGui::Image((ImTextureID)(uintptr_t)texID, size, ImVec2(0, 1), ImVec2(1, 0));
        ImGui::End();

        if (ImGui::BeginPopup("RenameObject")) {
            static char nameBuffer[128];
            static bool once = true;
        
            if (once && selected) {
                strcpy(nameBuffer, selected->GetName().c_str());
                once = false;
            }
        
            ImGui::InputText("##rename", nameBuffer, sizeof(nameBuffer));
            if (ImGui::IsKeyPressed(ImGuiKey_Enter) || ImGui::Button("OK")) {
                if (selected)
                    selected->SetName(nameBuffer);
                ImGui::CloseCurrentPopup();
                once = true;
            }
        
            ImGui::SameLine();
            if (ImGui::Button("Cancel")) {
                ImGui::CloseCurrentPopup();
                once = true;
            }
        
            ImGui::EndPopup();
        }

        // ImGui render
        ImGui::Render();
        int w, h;
        glfwGetFramebufferSize(window, &w, &h);
        glViewport(0, 0, w, h);
        glClear(GL_COLOR_BUFFER_BIT);
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());
        glfwSwapBuffers(window);
    }
}

void Engine::DrawObjectNode(const std::shared_ptr<Object> &obj)
{
    ImGuiTreeNodeFlags flags = ImGuiTreeNodeFlags_OpenOnArrow |
                               ImGuiTreeNodeFlags_SpanAvailWidth |
                               (obj == selected ? ImGuiTreeNodeFlags_Selected : 0);

    bool open = ImGui::TreeNodeEx((void *)(intptr_t)obj->id, flags, "%s", obj->GetName().c_str());

    if (ImGui::IsItemClicked())
        selected = obj;

    // === Context Menu on Object ===
    if (ImGui::BeginPopupContextItem())
    {
        if (ImGui::MenuItem("Rename"))
        {
            selected = obj;
            ImGui::OpenPopup("RenameObject");
        }
        if (ImGui::MenuItem("Delete"))
        {
            // Remove from parent or root
            if (obj->GetParent())
            {
                obj->GetParent()->RemoveChild(obj.get());
            }
            else
            {
                objects.erase(std::remove_if(objects.begin(), objects.end(),
                                             [&](const std::shared_ptr<Object> &o)
                                             { return o == obj; }),
                              objects.end());
            }
            ImGui::EndPopup();
            return; // Don't draw deleted node
        }
        if (ImGui::MenuItem("Create Child"))
        {
            auto child = std::make_shared<Object>("NewObject");
            obj->AddChild(child);
            selected = child;
            child->SetName("NewObject");
            ImGui::OpenPopup("RenameObject");
        }
        ImGui::EndPopup();
    }

    // === Drag and Drop ===
    if (ImGui::BeginDragDropSource())
    {
        ImGui::SetDragDropPayload("OBJECT", &obj, sizeof(obj));
        ImGui::Text("Move: %s", obj->GetName().c_str());
        ImGui::EndDragDropSource();
    }

    if (ImGui::BeginDragDropTarget())
    {
        if (const ImGuiPayload *payload = ImGui::AcceptDragDropPayload("OBJECT"))
        {
            auto dragged = *(std::shared_ptr<Object> *)payload->Data;
            if (dragged != obj)
                obj->AddChild(dragged);
        }
        ImGui::EndDragDropTarget();
    }

    // === Children ===
    if (open)
    {
        for (auto &child : obj->GetChildren())
            DrawObjectNode(child);
        ImGui::TreePop();
    }
}

void Engine::SaveScene(const std::string &path)
{
    YAML::Emitter out;
    YAML::Emitter sceneData;

    sceneData << YAML::BeginSeq;
    for (const auto &obj : objects)
        obj->Save(sceneData);
    sceneData << YAML::EndSeq;

    std::string sceneString = sceneData.c_str();
    unsigned char hash[SHA256_DIGEST_LENGTH];
    SHA256(reinterpret_cast<const unsigned char *>(sceneString.c_str()), sceneString.size(), hash);

    std::ostringstream hashHex;
    for (int i = 0; i < SHA256_DIGEST_LENGTH; ++i)
        hashHex << std::hex << std::setw(2) << std::setfill('0') << static_cast<int>(hash[i]);

    // Output full file
    out << YAML::BeginMap;
    out << YAML::Key << "engine_version" << YAML::Value << "0.1.0";
    out << YAML::Key << "scene_name" << YAML::Value << std::filesystem::path(path).stem().string();
    out << YAML::Key << "scene_hash" << YAML::Value << hashHex.str();
    out << YAML::Key << "format_version" << YAML::Value << 1;
    out << YAML::Key << "objects" << YAML::Value << YAML::Load(sceneString);
    out << YAML::EndMap;

    std::ofstream file(path);
    file << out.c_str();
}

bool VerifySceneHash(const YAML::Node &root)
{
    if (!root["scene_hash"] || !root["objects"])
        return false;

    YAML::Emitter sceneOnly;
    sceneOnly << root["objects"];

    std::string sceneString = sceneOnly.c_str();
    unsigned char hash[SHA256_DIGEST_LENGTH];
    SHA256(reinterpret_cast<const unsigned char *>(sceneString.c_str()), sceneString.size(), hash);

    std::ostringstream hashHex;
    for (int i = 0; i < SHA256_DIGEST_LENGTH; ++i)
        hashHex << std::hex << std::setw(2) << std::setfill('0') << static_cast<int>(hash[i]);

    return hashHex.str() == root["scene_hash"].as<std::string>();
}

void Engine::LoadScene(const std::string &path)
{
    YAML::Node root = YAML::LoadFile(path);

    if (!root["engine_version"] || !root["format_version"] || !root["scene_name"])
    {
        std::cerr << "[LoadScene] Missing required metadata!\n";
        return;
    }

    if (root["engine_version"].as<std::string>() != "0.1.0")
    {
        std::cerr << "[LoadScene] Version mismatch! Expected 0.1.0, got " << root["engine_version"].as<std::string>() << "\n";
        return;
    }

    if (!VerifySceneHash(root))
    {
        std::cerr << "[LoadScene] Scene hash does not match! File may be corrupted or tampered.\n";
    }

    objects.clear();
    const auto &objectArray = root["objects"];
    for (const auto &node : objectArray)
    {
        auto obj = std::make_shared<Object>("[DefaultObject]");
        obj->Load(node);
        objects.push_back(obj);
    }

    std::cout << "[LoadScene] Loaded scene: " << root["scene_name"].as<std::string>() << "\n";
}

void Engine::Shutdown()
{
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();
    glfwDestroyWindow(window);
    glfwTerminate();
}
