#include "Object.h"
#include "../Components/SpriteComponent.h"
#include "../Components/Component.h"


#include <algorithm>

static int nextID = 0;

Object::Object(const std::string& name)
    : name(name), localPosition(0.0f, 0.0f), id(nextID++), layer(1) {}

Object::~Object() {}

void Object::SetParent(Object* newParent) {
    if (parent)
        parent->RemoveChild(this);
    parent = newParent;
}

void Object::AddChild(std::shared_ptr<Object> child) {
    child->SetParent(this);
    children.push_back(child);
}

void Object::RemoveChild(Object* child) {
    children.erase(std::remove_if(children.begin(), children.end(),
        [child](const std::shared_ptr<Object>& obj) {
            return obj.get() == child;
        }), children.end());
}

glm::vec2 Object::GetLocalPosition() const {
    return localPosition;
}

void Object::SetLocalPosition(glm::vec2 pos) {
    localPosition = pos;
}

glm::vec2 Object::GetWorldPosition() const {
    if (parent)
        return parent->GetWorldPosition() + localPosition;
    return localPosition;
}

const std::string& Object::GetName() const { return name; }
void Object::SetName(const std::string& n) { name = n; }
std::vector<std::shared_ptr<Object>>& Object::GetChildren() { return children; }
Object* Object::GetParent() const { return parent; }

// === Component System ===

template<typename T>
std::shared_ptr<T> Object::GetComponent() const {
    for (const auto& comp : components) {
        if (auto casted = std::dynamic_pointer_cast<T>(comp))
            return casted;
    }
    return nullptr;
}

template<typename T>
std::shared_ptr<T> Object::AddComponent() {
    auto existing = GetComponent<T>();
    if (existing) return existing;

    std::shared_ptr<T> component = std::make_shared<T>(this);
    components.push_back(component);
    return component;
}

template<typename T>
void Object::RemoveComponent() {
    components.erase(std::remove_if(components.begin(), components.end(),
        [](const std::shared_ptr<Component>& comp) {
            return std::dynamic_pointer_cast<T>(comp) != nullptr;
        }), components.end());
}

// Explicit template instantiation
template std::shared_ptr<class SpriteComponent> Object::GetComponent<class SpriteComponent>() const;
template std::shared_ptr<class SpriteComponent> Object::AddComponent<class SpriteComponent>();
template void Object::RemoveComponent<class SpriteComponent>();

void Object::Save(YAML::Emitter& out) const {
    out << YAML::BeginMap;
    out << YAML::Key << "name" << YAML::Value << name;
    out << YAML::Key << "position" << YAML::Value << YAML::Flow << YAML::BeginSeq << localPosition.x << localPosition.y << YAML::EndSeq;

    out << YAML::Key << "components" << YAML::Value << YAML::BeginSeq;
    for (const auto& comp : components) {
        comp->Save(out);
    }
    out << YAML::EndSeq;

    out << YAML::Key << "children" << YAML::Value << YAML::BeginSeq;
    for (const auto& child : children)
        child->Save(out);
    out << YAML::EndSeq;

    out << YAML::EndMap;
}

void Object::Load(const YAML::Node& node) {
    name = node["name"].as<std::string>();
    auto pos = node["position"];
    localPosition = { pos[0].as<float>(), pos[1].as<float>() };

    components.clear();
    if (node["components"]) {
        for (const auto& compNode : node["components"]) {
            std::string type = compNode["type"].as<std::string>();
            if (type == "SpriteComponent") {
                auto comp = AddComponent<class SpriteComponent>();
                comp->Load(compNode);
            }
        }
    }

    children.clear();
    if (node["children"]) {
        for (const auto& childNode : node["children"]) {
            auto child = std::make_shared<Object>("(loaded)");
            child->Load(childNode);
            AddChild(child);
        }
    }
}
