#include "Shader.h"
#include <fstream>
#include <sstream>
#include <iostream>

Shader::Shader() {}
Shader::~Shader() { Clear(); }

void Shader::Clear() {
    if (id != 0) {
        glDeleteProgram(id);
        id = 0;
    }
    uniformCache.clear();
}

std::string Shader::ReadFile(const std::string& path) {
    std::ifstream stream(path);
    std::stringstream buffer;
    buffer << stream.rdbuf();
    return buffer.str();
}

GLuint Shader::Compile(GLenum type, const std::string& source) {
    GLuint shader = glCreateShader(type);
    const char* src = source.c_str();
    glShaderSource(shader, 1, &src, nullptr);
    glCompileShader(shader);

    GLint success;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &success);
    if (!success) {
        char info[512];
        glGetShaderInfoLog(shader, 512, nullptr, info);
        std::cerr << "[Shader Compile Error]:\n" << info << std::endl;
    }

    return shader;
}

bool Shader::LoadFromFile(const std::string& vertexPath, const std::string& fragmentPath) {
    return LoadFromSource(ReadFile(vertexPath), ReadFile(fragmentPath));
}

bool Shader::LoadFromSource(const std::string& vertexSrc, const std::string& fragmentSrc) {
    Clear();

    GLuint vertex = Compile(GL_VERTEX_SHADER, vertexSrc);
    GLuint fragment = Compile(GL_FRAGMENT_SHADER, fragmentSrc);

    id = glCreateProgram();
    glAttachShader(id, vertex);
    glAttachShader(id, fragment);
    glLinkProgram(id);

    glDeleteShader(vertex);
    glDeleteShader(fragment);

    GLint success;
    glGetProgramiv(id, GL_LINK_STATUS, &success);
    if (!success) {
        char info[512];
        glGetProgramInfoLog(id, 512, nullptr, info);
        std::cerr << "[Shader Link Error]:\n" << info << std::endl;
        return false;
    }

    return true;
}

void Shader::Use() const {
    glUseProgram(id);
}

GLuint Shader::GetID() const {
    return id;
}

GLint Shader::GetUniformLocation(const std::string& name) {
    if (uniformCache.contains(name)) return uniformCache[name];
    GLint loc = glGetUniformLocation(id, name.c_str());
    if (loc == -1)
        std::cerr << "[Shader] Warning: Uniform not found: " << name << "\n";
    uniformCache[name] = loc;
    return loc;
}

// === Uniform Setters ===

void Shader::SetBool(const std::string& name, bool value) {
    glUniform1i(GetUniformLocation(name), (int)value);
}

void Shader::SetInt(const std::string& name, int value) {
    glUniform1i(GetUniformLocation(name), value);
}

void Shader::SetFloat(const std::string& name, float value) {
    glUniform1f(GetUniformLocation(name), value);
}

void Shader::SetVec2(const std::string& name, const glm::vec2& value) {
    glUniform2fv(GetUniformLocation(name), 1, &value[0]);
}

void Shader::SetVec3(const std::string& name, const glm::vec3& value) {
    glUniform3fv(GetUniformLocation(name), 1, &value[0]);
}

void Shader::SetVec4(const std::string& name, const glm::vec4& value) {
    glUniform4fv(GetUniformLocation(name), 1, &value[0]);
}

void Shader::SetMat4(const std::string& name, const glm::mat4& mat) {
    glUniformMatrix4fv(GetUniformLocation(name), 1, GL_FALSE, &mat[0][0]);
}
